/*  kloc.cpp - source code statistics program
  
    Copyright (C) 2001 dinker charak

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The contact addresses for author is as follows:
    Web:   http://sarovar.org/projects/kloc/
    Email: ddiinnxx at users.sarovar.org
*/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>

#include <libgen.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include "GetOpt.h"

#define PROD_VERSION         VERSION
#define DEFAULT_OUT_NAME     "kloc.cfg"

#define LABEL_PROCESS_FILE  "Process File"
#define LABEL_PROCESS_EXT   "Process Extension"
#define LABEL_IGNORE_FILE   "Ignore File"
#define LABEL_IGNORE_EXT    "Ignore Extension"

#define REV_PREFIX          "$Revision: "

#define LINE_EMPTY            1
#define LINE_COMMENT          2
#define LINE_EMBEDDED_COMMENT 3
#define LINE_CODE             4
#define LINE_PROCESS_EXT      5
#define LINE_PROCESS_FILE     6
#define LINE_IGNORE_EXT       7
#define LINE_IGNORE_FILE      8

#define DEFAULT_CWD_SIZE      40

#define FLAG_RESET_ALL        0x00000000 
#define FLAG_ONLY_CUMULATIVE  0x00000001 
#define FLAG_VERBOSE          0x00000002 
#define FLAG_NOT_RECURSIVE    0x00000004 
#define FLAG_MORE_STATS       0x00000008 
#define FLAG_USE_CFG_FILE     0x00000010 
#define FLAG_USE_ARGS         0x00000020 
#define FLAG_USE_CWD          0x00000040 
#define FLAG_IGNORE_HIDDEN    0x00000080 
#define FLAG_NO_FLAG          0x80000000 

#define FLAG_PROCESS_FILE     0x00100000 
#define FLAG_PROCESS_EXT      0x00200000 
#define FLAG_IGNORE_FILE      0x00400000 
#define FLAG_IGNORE_EXT       0x00800000 

void usage(int);
void more_usage(int);
void generate_cfg(int);
void process_cfg(int &);
void validateFlags(int &);
int  getLineType(std::string &);
int  getCfgType(std::string &);
void processDirs(int);
void processFiles(std::string, int &);
void parse_dirs(std::string, int);
bool fileCfgStatus(std::string, int);
void klocfile(std::string, int);
bool hasNothing(std::string, int);
std::string getFileExt(std::string, int);
void reportErr(std::string, std::string, int, int);

static unsigned long int g_empty, g_comment, g_embedded, g_code, 
                         g_total, g_files;

static std::vector<std::string> pro_ext;
static std::vector<std::string> pro_fil;
static std::vector<std::string> ign_ext;
static std::vector<std::string> ign_fil;

int main(int argc, char *argv[])
{
	if (argc == 1)
	{
		usage(0);
		exit(0);
	}
	
	int  opt, flags;
	std::string ifileName;
	
	flags &= FLAG_RESET_ALL;
	flags = FLAG_NO_FLAG;
	
	otf::GetOpt getopt(argc, argv, "vHhcugfsi:da");
	while ((opt = getopt()) != -1)
	{
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "kloc: opt being processed is " << opt << std::endl;
		}

		switch (opt) 
		{
		case 'v':
			flags |= FLAG_VERBOSE;
			break;
		case 'H':
			more_usage(flags);
			exit (0);
		case 'h':
			usage(flags);
			exit (0);
		case 'c':
			generate_cfg(flags);
			exit (0);
			break;
		case 'u':
			flags |= FLAG_USE_CFG_FILE;
			flags &= ~FLAG_NO_FLAG;
			break;
		case 'g':
			flags |= FLAG_ONLY_CUMULATIVE;
			break;
		case 'f':
			flags |= FLAG_NOT_RECURSIVE;
			break;
		case 's':
			flags |= FLAG_MORE_STATS;
			break;
		case 'i':
			flags |= FLAG_USE_ARGS;
			flags &= ~FLAG_NO_FLAG;
				ifileName = getopt.arg();
			break;
		case 'd':
			flags |= FLAG_USE_CWD;
			flags &= ~FLAG_NO_FLAG;
			break;
		case 'a':
			flags |= FLAG_IGNORE_HIDDEN;
			break;
		case '?':
		default:
			usage(0);
			exit (0);
		} // switch (opt) 
	} // while ((opt = getopt()) != -1)
	
	if (flags & FLAG_NO_FLAG)
	{
		std::string err = "Use -i, -u or -d to specify input file information. Try 'kloc -h' for more information.";
		reportErr("main", err, -1, flags);
	}

	if (flags & FLAG_USE_CFG_FILE)
	{
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "kloc: looking for " << DEFAULT_OUT_NAME << std::endl;
		}
		
		process_cfg(flags);
	}
	
	validateFlags(flags);

	if (flags & FLAG_USE_ARGS)
	{
		processFiles(ifileName, flags);
	}

	if (flags & FLAG_USE_CWD)
	{
		processDirs(flags);
	}
	
	std::cout << std::endl << "Cumulative stats" << std::endl;
	std::cout << "Total Empty Lines	      = " << g_empty << std::endl;
	std::cout << "Total Comments          = " << g_comment << std::endl;
	std::cout << "Total Embedded Comments = " << g_embedded << std::endl;
	std::cout << "Total Code              = " << g_code << std::endl;
	std::cout << "Total Lines	      = " << g_total  << std::endl;
	std::cout << "Total Files	      = " << g_files << std::endl;
	std::cout << std::endl;

	if (flags & FLAG_MORE_STATS)
	{
		float stuff = (((float) g_code) / ((float) g_total)) * 100.0;
		std::cout << stuff << "% of written lines were code" << std::endl;

		float comments = (((float) g_comment + (float) g_embedded) / ((float) g_code)) * 100.0;
		std::cout << comments << "% of code was commented" << std::endl;

		float codecomments = (((float) g_comment + (float) g_embedded + (float) g_code) / ((float) g_total)) * 100.0;
		std::cout << codecomments << "% of written lines were code or comments" << std::endl;
	}
	
	return 0;
}

void processDirs(int flags)
{
	char *buf;
	int  size = DEFAULT_CWD_SIZE;
	
	buf = new char [size];
	
	while (getcwd(buf, size) == NULL)
	{
		if (errno == ERANGE)
		{
			delete buf;
			size+=DEFAULT_CWD_SIZE;
			buf = new char [size];
		}
		else
		{
			std::string err = "Error ("; err += errno; err += ") while getting cwd";
			reportErr("processDirs", err, -1, flags);
		}
	}
	
	std::string path(buf);
	delete buf;

	if (flags & FLAG_NOT_RECURSIVE)
	{
		std::cout << "parsing " << path << std::endl;
	}
	else
	{
		std::cout << "parsing recursively " << path << std::endl;
	}
	     
	parse_dirs(path, flags);   
}

void parse_dirs(std::string path, int flags)
{
	DIR *dirp, *dirn;
	struct dirent* dp;

	if (path.empty()) 
	{
		std::string err = "Path is empty";
		reportErr("parse_dirs", err, -1, flags);
	}

	dirp = opendir(path.c_str());
	
	if (dirp == NULL) 
	{
		std::string err = "opendir() returned a NULL";
		reportErr("parse_dirs", err, -1, flags);
	}
  
	while ((dp = readdir(dirp)) != NULL) 
	{
		if ( ! ( ( ! strcmp(".", dp->d_name)) || ( ! strcmp("..", dp->d_name)) || ( (flags & FLAG_IGNORE_HIDDEN) && ( dp->d_name[0] == '.') ) ) )
		{
			std::string thefile = path + "/" + dp->d_name;
		
			if (flags & FLAG_VERBOSE)
			{
				std::cout << std::endl << "parse_dirs: processing " << thefile << std::endl;
			}
		
			dirn = opendir(thefile.c_str()); 

			// check if a dir
			if (dirn != NULL) 
			{
				if (flags & FLAG_VERBOSE)
				{
					std::cout << "parse_dirs: changing dir to " << thefile << std::endl;
				}

				// check contents of the dir
				if ( !(flags & FLAG_NOT_RECURSIVE) )
				{
					parse_dirs(thefile, flags);
				}

				if (flags & FLAG_VERBOSE)
				{
					std::cout << "parse_dirs: closed access to " << thefile << std::endl;
				}

				// moved closedir(dirn) here as sending NULL to closedir() causes seg fault in solaris.
				// also added closedir() return status check.

				// close the desriptor. Added as fix to EMFILE (24) in case of large dir heirarchy.
				if ( closedir(dirn) == -1)
				{
					std::string err = "Cound not close "; err+= thefile; err+= ".";
					reportErr("parse_dirs", err, 0, flags);
				}
				else 
				{
					if (flags & FLAG_VERBOSE)
					{
						std::cout << "parse_dirs: closed access to " << thefile << std::endl;
					}
		    		}
			}
			else // if (dirn != NULL) 
			{
				// is a file, process it
				// see if file needs to be processed if we are using cgf
				if (flags & FLAG_USE_CFG_FILE)
				{
					bool doProcess;
					
					doProcess = fileCfgStatus(thefile, flags);
					
					if (doProcess)
					{
						if (flags & FLAG_VERBOSE)
						{
							std::cout << "parse_dirs: parsing " << thefile;
						}	
						
						// process all files if not using cgf
						klocfile(thefile, flags);
					} 
					else // if (doProcess)
					{
						if (flags & FLAG_VERBOSE)
						{	
							std::cout << "parse_dirs: ignoring " << thefile;
						}
					} // if (doProcess)
				}
				else // if (flags & FLAG_USE_CFG_FILE)
				{
					if (flags & FLAG_VERBOSE)
					{
						std::cout << "parse_dirs: parsing " << thefile;
					}

					// process all files if not using cgf
					klocfile(thefile, flags);
				} // if (flags & FLAG_USE_CFG_FILE)      
			} // if (dirn != NULL)
		} // if ( ! ( ( ! strcmp(".", dp->d_name)) || ( ! strcmp("..", dp->d_name)) ...
	} // while ((dp = readdir(dirp)) != NULL)

	// close the desriptor
	if ( closedir(dirp) == -1)
	{
		std::string err = "Cound not close "; err+= path.c_str(); err+= ".";
		reportErr("parse_dirs", err, 0, flags);
	}
	else if (flags & FLAG_VERBOSE)
	{
		std::cout << "parse_dirs: closed access to " << path.c_str() << std::endl;
	}
}

void klocfile(std::string infilename, int flags)
{
	unsigned long int l_empty, l_comment, l_embedded, l_code, l_total;
	int type;
	std::string ln;

#ifdef ROSPELL
	std::string ext = getFileExt(infilename, flags);
	if (ext.empty())
		return;
	if (ext != "c" && ext != "h" && ext != "cpp" && ext != "cc" && ext != "java" &&
	    ext != "js" && ext != "cs" && ext != "ph3" && ext != "ph4" && ext != "php" && ext != "phps")
		return;
#endif
	std::ifstream infile (infilename.c_str());

	if (infile == NULL)
	{
		std::string err = "Error opening input file - "; err+= infilename; err+= ". Ignoring it!";
		reportErr("klocfile", err, 0, flags);

		if (flags & FLAG_USE_ARGS)
		{
			std::string err = "Make sure -i is followed by a valid filename only.";
			reportErr("klocfile", err, -1, flags);
		}
		
		return;
	} // if (infile == NULL)

	l_empty    = 0;
	l_comment  = 0;
	l_embedded = 0;
	l_code     = 0;
	l_total    = 0;
	
	while (getline(infile, ln))
	{
		type = getLineType(ln);
		switch (type)
		{
			case LINE_EMPTY	     : l_empty++	      ; break;
			case LINE_COMMENT	   : l_comment++	    ; break;
			case LINE_EMBEDDED_COMMENT : l_embedded++; l_code++; break;
			case LINE_CODE	      : l_code++		; break;
		}
		
		l_total++;
	}	
	
	if ( !(flags & FLAG_ONLY_CUMULATIVE) )
	{
		std::cout << std::endl;
#ifdef ROSPELL
		std::cout << infilename << ":1" << std::endl;
#else
		std::cout << "File		= " << infilename << std::endl;
#endif
		std::cout << "Empty Lines	= " << l_empty << std::endl;
		std::cout << "Comments	   = " << l_comment << std::endl;
		std::cout << "Embedded Comments = " << l_embedded << std::endl;
		std::cout << "Code		= " << l_code << std::endl;
		std::cout << "Total Lines	= " << l_total << std::endl;
	}
	
	g_empty    +=l_empty;
	g_comment  +=l_comment;
	g_embedded +=l_embedded;
	g_code     +=l_code;
	g_total    +=l_total;
	g_files    ++;
	return;	
}

void generate_cfg(int flags)
{
	std::string outfilename;
	std::ostringstream version_holder;
	
	version_holder << PROD_VERSION;

	outfilename  = DEFAULT_OUT_NAME;
	
	std::ofstream outfile (outfilename.c_str());

	if (outfile == NULL)
	{
		std::string err = "Error opening "; err += outfilename;
		reportErr("generate_cfg", err, -1, flags);
	}

	std::string msg, time_str;

	msg =  "# Template for this was automatically generated by kloc utility\n";
	msg +=  "# ";
	msg += REV_PREFIX;
	msg += version_holder.str();
	msg += " ";
	msg += "on ";

	time_t rawtime;
	struct tm * timeinfo;

	time ( &rawtime );
	timeinfo = localtime ( &rawtime );
	time_str = asctime(timeinfo);

	// removing \n at end
	time_str = time_str.substr(0, time_str.length()-1);

	msg += time_str;
	msg += "\n";
	
	outfile << msg << std::endl;
	outfile << "# Do kloc -H on how to use this file" << std::endl;
	outfile << "# " << std::endl;
	outfile << "" << std::endl;
	outfile << "[Process File]" << std::endl;
	outfile << "" << std::endl;
	outfile << "[Process Extension]" << std::endl;
	outfile << "" << std::endl;
	outfile << "[Ignore File]" << std::endl;
	outfile << "" << std::endl;
	outfile << "[Ignore Extension]" << std::endl;
	outfile << "" << std::endl;
	 
	return;	
}

int getLineType(std::string &tmpStr)
{
	std::string token;
	std::string::size_type pos;
	static bool foundStartComment=false;

	// remove leading space
	while (1)
	{
		pos = tmpStr.find(' ');
		if (pos == 0)
		{
			tmpStr = tmpStr.substr(1, tmpStr.length());
		}
		else
		{
			break;
		}
	}

	// remove leading tabs
	while (1)
	{
		pos = tmpStr.find('\t');
		if (pos == 0)
		{
			tmpStr = tmpStr.substr(1, tmpStr.length());
		}
		else
		{
			break;
		}
	}

	if (tmpStr.length() == 0)
	{
		return LINE_EMPTY;
	}

	// Support for Win/Cygwin, files with CR/LF
	if(tmpStr[0]<'\x20')
	{
		return LINE_EMPTY;
	}

	// find the end of word
	pos = tmpStr.find(' ');

	// read the first word
	token = tmpStr.substr(0, pos);

	// If we're inside a comment section /* */
	// check for end mark */
	if (foundStartComment)
	{
		pos = tmpStr.find("*/");
		
		if (pos != std::string::npos)
		{
			foundStartComment = false;
		}
		else
		{
			foundStartComment = true;
		}
	      
		return LINE_COMMENT;
	}

	if ( ( (token[0] == '/') && (token[1] == '/') ) || (token[0] == '#') )
	{
		return LINE_COMMENT;
	}

	if ( (token[0] == '/') && (token[1] == '*'))
	{
		tmpStr = tmpStr.substr(2, tmpStr.length());
	 
		// Check if it ends on the same line
		pos = tmpStr.find("*/");
	 
		if(pos != std::string::npos)
		{
			foundStartComment = false;
		}
		else
		{
			foundStartComment = true;
		}
	 
		return LINE_COMMENT;
	}

	// search for // in the line
	pos = tmpStr.find("//");
	
	if (pos != std::string::npos)
	{
		return LINE_EMBEDDED_COMMENT;
	}

	// search for /* in the line
	pos = tmpStr.find("/*");
	
	if (pos != std::string::npos)
	{
		tmpStr = tmpStr.substr(2, tmpStr.length());
	 
		// Check if it ends on the same line
		pos = tmpStr.find("*/");
	 
		if(pos != std::string::npos)
		{
			foundStartComment = false;
		}
		else
		{
			foundStartComment = true;
		}
	 	
		return LINE_EMBEDDED_COMMENT;
	}

	return LINE_CODE;
}

int getCfgType(std::string &tmpStr)
{
	std::string token;
	std::string::size_type pos;
	
	// search for LABEL_PROCESS_EXT in the line
	pos = tmpStr.find(LABEL_PROCESS_EXT);
	if (pos != std::string::npos)
	{
		return LINE_PROCESS_EXT;
	}
	
	// search for LABEL_PROCESS_FILE in the line
	pos = tmpStr.find(LABEL_PROCESS_FILE);
	if (pos != std::string::npos)
	{
		return LINE_PROCESS_FILE;
	}

	// search for LABEL_IGNORE_EXT in the line
	pos = tmpStr.find(LABEL_IGNORE_EXT);
	if (pos != std::string::npos)
	{
		return LINE_IGNORE_EXT;
	}

	// search for LABEL_IGNORE_FILE in the line
	pos = tmpStr.find(LABEL_IGNORE_FILE);
	if (pos != std::string::npos)
	{
		return LINE_IGNORE_FILE;
	}
	
	return LINE_CODE;
}

void process_cfg(int &flags)
{
	std::string ln;

	std::ifstream infile (DEFAULT_OUT_NAME);

	if (infile == NULL)
	{
		std::string err = "Error opening cfg file - "; err += DEFAULT_OUT_NAME;
		reportErr("process_cfg", err, 0, flags);
		err = "Generate one using -c option. Try 'kloc -h' for more information.";
		reportErr("process_cfg", err, -1, flags);
	}

	if (flags & FLAG_VERBOSE)
	{
		std::cout << "process_cfg: parsing the cfg file" << std::endl;
	}

	bool notValidated = true;
	
	while (getline(infile, ln))
	{
		int lnType, cfgType;
		static int configuring_now;
		
		lnType = getLineType(ln);
		
		if ( notValidated && (lnType == LINE_COMMENT) )
		{
			std::string::size_type pos;
			
			pos = ln.find(REV_PREFIX);
			
			if (pos != std::string::npos)
			{
				std::string prefixStr(REV_PREFIX);
				std::ostringstream version_holder;
				
				std::string revCfg = ln.substr(pos + prefixStr.length(), ln.length());
				
				std::string::size_type revisionPos = revCfg.find(" ");
				revCfg = revCfg.substr(0, revisionPos);
				
				if (flags & FLAG_VERBOSE)
				{
					std::cout << "process_cfg: ver of cfg file is " << revCfg << std::endl;
				}
				
				version_holder << PROD_VERSION;
				std::string revPrg = version_holder.str();;
				
				if (flags & FLAG_VERBOSE)
				{
					std::cout << "process_cfg: ver of kloc is " << revPrg << std::endl;
				}
				
				float revCfgNum = strtod(revCfg.c_str(), (char **)NULL);
				
				if ( (errno == ERANGE) || (errno == EINVAL) )
				{
			      		std::string err = "Error converting cfg file ver to integer. Read - "; err += revCfg;
					reportErr("process_cfg", err, -1, flags);
				}
				
				float revPrgNum = strtod(revPrg.c_str(), (char **)NULL);
				
				if ( (errno == ERANGE) || (errno == EINVAL) )
				{
			      		std::string err = "Error converting kloc ver to integer. Read - "; err += revPrg;
					reportErr("process_cfg", err, -1, flags);
				}

				if (revCfgNum != revPrgNum)
				{
					if (revPrgNum > revCfgNum)
					{
						std::string err = "The cfg file was generated by an older version and may not work with this version.";
						reportErr("process_cfg", err, 0, flags);
					}
					else
					{
						std::string err = "The cfg file was generated by an newer version and may not work with this version.";
						reportErr("process_cfg", err, 0, flags);
					}
					
					std::string err = "Generate another one using -c option. Try 'kloc -h' for more information";
					reportErr("process_cfg", err, -1, flags);
				}
				
				notValidated = false;
				
			} // if (pos != std::string::npos)
		} // if ( notValidated && (lnType == LINE_COMMENT) )
		
		if ( (lnType != LINE_EMPTY) && (lnType != LINE_COMMENT) )
		{
			if (notValidated)
			{
				std::string err = "Header for cfg file seems to be missing or corrupted. Generate another one using -c option. Try 'kloc -h' for more information";
				reportErr("process_cfg", err, -1, flags);
			}			
			
			cfgType = getCfgType(ln);

			switch (cfgType)
			{
				case LINE_PROCESS_EXT   : 
					if (flags & FLAG_VERBOSE)
					{
						std::cout << "process_cfg: reading " << LABEL_PROCESS_EXT << std::endl;
					}
					configuring_now = LINE_PROCESS_EXT;
					break;
				case LINE_PROCESS_FILE  : 
					if (flags & FLAG_VERBOSE)
					{
						std::cout << "process_cfg: reading " << LABEL_PROCESS_FILE << std::endl;
					}
					configuring_now = LINE_PROCESS_FILE;
					break;
				case LINE_IGNORE_EXT 	: 
					if (flags & FLAG_VERBOSE)
					{
						std::cout << "process_cfg: reading " << LABEL_IGNORE_EXT << std::endl;
					}
					configuring_now = LINE_IGNORE_EXT;
					break;
				case LINE_IGNORE_FILE   : 
					if (flags & FLAG_VERBOSE)
					{
						std::cout << "process_cfg: reading " << LABEL_IGNORE_FILE << std::endl;
					}
					configuring_now = LINE_IGNORE_FILE;
					break;
				default			:
					switch (configuring_now)
					{
						case LINE_PROCESS_EXT   : 
							if (! hasNothing(ln, flags) )
							{
								if (flags & FLAG_VERBOSE)
								{
									std::cout << "process_cfg: adding '" << ln << "' from " << LABEL_PROCESS_EXT << std::endl;
								}
								pro_ext.push_back(ln);
								flags |= FLAG_PROCESS_EXT;
							}	
							break;
						case LINE_PROCESS_FILE  : 
							if (! hasNothing(ln, flags) )
							{
								if (flags & FLAG_VERBOSE)
								{
									std::cout << "process_cfg: adding '" << ln << "' from " << LABEL_PROCESS_FILE << std::endl;
								}
								pro_fil.push_back(ln);
								flags |= FLAG_PROCESS_FILE;
							}
							break;
						case LINE_IGNORE_EXT 	: 
							if (! hasNothing(ln, flags) )
							{
								if (flags & FLAG_VERBOSE)
								{
									std::cout << "process_cfg: adding '" << ln << "' from " << LABEL_IGNORE_EXT << std::endl;
								}
								ign_ext.push_back(ln);
								flags |= FLAG_IGNORE_EXT;
							}
							break;
						case LINE_IGNORE_FILE   : 
							if (! hasNothing(ln, flags) )
							{
								if (flags & FLAG_VERBOSE)
								{
									std::cout << "process_cfg: adding '" << ln << "' from " << LABEL_IGNORE_FILE << std::endl;
								}
								ign_fil.push_back(ln);
								flags |= FLAG_IGNORE_FILE;
							}
							break;
						default			:
							std::string err = "While adding '"; err += ln; err += "', strange type being configured - ";
							err += configuring_now;
							reportErr("process_cfg", err, -1, flags);

					} // switch (configuring_now)
			} // switch (cfgType)
		} // if ( (lnType != LINE_EMPTY) && (lnType != LINE_COMMENT) )
	} // while (getline(infile, ln))
	
	return;
}
 
bool fileCfgStatus(std::string thefile, int flags)
{
	std::string fname;
	std::string ext;
	
	fname = basename((char *)thefile.c_str());

	if (flags & FLAG_VERBOSE)
	{
		std::cout << "fileCfgStatus: checking for file " << fname << " in configuration file" << std::endl;
	}
	
	//	 if    |    then
	// ------------------------
	//      PF   PE  |  IF   IE
	// i    1    1   |  x    0
	if ( (flags & FLAG_PROCESS_FILE) && (flags & FLAG_PROCESS_EXT) )
	{
		bool process = false;
		ext = getFileExt(fname, flags);
		std::vector<std::string>::const_iterator itr;
		
		// see if the filename is mentioned in process ext
		for (itr = pro_ext.begin(); itr != pro_ext.end(); itr++)
		{
			if (*itr == ext)
			{
				// file mentioned ... we will process
				process = true;
			}
		}
		
		// if not mentioned in process ext, see if mentioned in process file
		if (!process)
		{
			for (itr = pro_fil.begin(); itr != pro_fil.end(); itr++)
			{
				if (*itr == fname)
				{
					// file mentioned ... we will process
					process = true;
				}
			}
		}
		
		// if we are about to process, see if mentioned in ignore files ... like ignore one file from this ext
		if (process)
		{
			for (itr = ign_fil.begin(); itr != ign_fil.end(); itr++)
			{
				if (*itr == fname)
				{
					// found it ... we are ignoring it
					process = false;
				}
			}
		}
		
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "fileCfgStatus: '" << thefile << "' process status is: " << process << std::endl;
		}

		return process;
	}

	//	 if    |    then
	// ------------------------
	//      PF   PE  |  IF   IE
	// ii   1    0   |  0    0
	if ( (flags & FLAG_PROCESS_FILE) && (flags & ~FLAG_PROCESS_EXT) )
	{
		bool process = false;
		std::vector<std::string>::const_iterator itr;
		
		// see if the filename is mentioned in process files
		for (itr = pro_fil.begin(); itr != pro_fil.end(); itr++)
		{
			if (*itr == fname)
			{
				// found it .. we will process
				process = true;
			}
		}
		
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "fileCfgStatus: '" << thefile << "' process status is: " << process << std::endl;
		}

		return process;
	}
	
	//	 if    |    then
	// ------------------------
	//      PF   PE  |  IF   IE
	// iii  0    1   |  x    0
	if ( (flags & ~FLAG_PROCESS_FILE) && (flags & FLAG_PROCESS_EXT) )
	{
		bool process = false;
		ext = getFileExt(fname, flags);
		std::vector<std::string>::const_iterator itr;
		
		// check if ext mentioned in process ext
		for (itr = pro_ext.begin(); itr != pro_ext.end(); itr++)
		{
			if (*itr == ext)
			{
				// found it ... we will process
				process = true;
			}
		}
		
		// if processing, check if some file is in ignore file list
		if (process)
		{
			for (itr = ign_fil.begin(); itr != ign_fil.end(); itr++)
			{
				if (*itr == fname)
				{
					// found it ... will not process now
					process = false;
				}
			}
		}
		
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "fileCfgStatus: '" << thefile << "' process status is: " << process << std::endl;
		}

		return process;
	}

	//	 if    |    then
	// ------------------------
	//      PF   PE  |  IF   IE
	// iv   0    0   |  x    x
	if ( (flags & ~FLAG_PROCESS_FILE) && (flags & ~FLAG_PROCESS_EXT) )
	{
		bool process = true;
		ext = getFileExt(fname, flags);
		std::vector<std::string>::const_iterator itr;
		
		// see if the filename is mentioned in ignore files
		for (itr = ign_fil.begin(); itr != ign_fil.end(); itr++)
		{
			if (*itr == fname)
			{
				// found it ... so we will ignore
				process = false;
			}
		}
		
		// if still not found, see if present in ignore file
		if (process)
		{
			for (itr = ign_ext.begin(); itr != ign_ext.end(); itr++)
			{
				if (*itr == ext)
				{
					// found it ... so we will ignore
					process = false;
				}
			}
		}
		
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "fileCfgStatus: '" << thefile << "' process status is: " << process << std::endl;
		}

		return process;
	}

	if (flags & FLAG_VERBOSE)
	{
		std::cout << "fileCfgStatus: '" << thefile << "' process status is: " << true << std::endl;
	}

	return true;
}

std::string getFileExt(std::string fname, int flags)
{
	std::string ext;
	
	// length is a 1 based count whereas in below loop
	// we run a 0 based count
	unsigned int counter = fname.length() - 1;

	while (counter > 0)
	{
		if (fname[counter] == '.')
		{
			ext = fname.substr(counter+1, fname.length() - counter);
			break;
		}

		counter--;
	}

	if (flags & FLAG_VERBOSE)
	{
		if (ext.empty())
		{
			std::cout << "getFileExt: '" << fname << "' has no extension" << ext << std::endl;
		}
		else
		{
			std::cout << "getFileExt: extension of " << fname << " is " << ext << std::endl;
		}
	}

	return ext;
}

bool hasNothing(std::string ln, int flags)
{
	bool ret = true;
	
	int  i,
	     len = ln.length();
	
	if (len < 0)
	{
		std::string err = "Length of entry is unexpectedly long.";
		reportErr("hasNothing", err, -1, flags);
	}
	
	if (!ln.empty())
	{
		for (i=0; i<len; i++)
		{
			if ( ! ( (ln[i] == ' ') || (ln[i] == '\t') || (ln[i] == '\0') ) )
			{
				ret = false;
				break;
			}
		}
	}
	
	return ret;
}

void validateFlags(int &flags)
{
	if (flags & FLAG_USE_CFG_FILE)
	{
		/* The rules truth table

			if    |    then
		------------------------
		     PF   PE  |  IF   IE
		i    1    1   |  x    0
		ii   1    0   |  0    0
		iii  0    1   |  x    0
		iv   0    0   |  x    x

		PF - Process File
		PE - Process Extension
		IF - Ignore File
		IE - Ignore Extension
		1  - Will be considered
		0  - Will not be considered even if entry present
		x  - Will depend if entry is present
		*/

		if ( (flags & FLAG_PROCESS_FILE) && (flags & FLAG_PROCESS_EXT) )
		{
			flags &= ~FLAG_IGNORE_EXT;
		}

		if ( (flags & FLAG_PROCESS_FILE) && (flags & ~FLAG_PROCESS_EXT) )
		{
			flags &= ~FLAG_IGNORE_FILE;
			flags &= ~FLAG_IGNORE_EXT;
		}

		if ( (flags & ~FLAG_PROCESS_FILE) && (flags & FLAG_PROCESS_EXT) )
		{
			flags &= ~FLAG_IGNORE_EXT;
		}

		if (flags & FLAG_VERBOSE)
		{
			printf("validateFlags: flags = 0x%x\n", flags);
			std::cout << "validateFlags: processing information from fields:" << std::endl;

			if (flags & FLAG_PROCESS_FILE)
			{
				std::cout << "validateFlags: Process File " << std::endl;
			}

			if (flags & FLAG_PROCESS_EXT)
			{
				std::cout << "validateFlags: Process Extension " << std::endl;
			}

			if (flags & FLAG_IGNORE_FILE)
			{
				std::cout << "validateFlags: Ignore File " << std::endl;
			}

			if (flags & FLAG_IGNORE_EXT)
			{
				std::cout << "validateFlags: Ignore Extension" << std::endl;
			}
		} // if (flags & FLAG_VERBOSE)
	} // if (flags & FLAG_USE_CFG_FILE)

	if (flags & FLAG_USE_ARGS)
	{
		if (flags & FLAG_USE_CWD)
		{
			flags &= ~FLAG_USE_CWD;
      			std::cout << "'-d' is being overridden by '-i'." << std::endl;
		}

		if (flags & FLAG_USE_CFG_FILE)
		{
			// reset use cfg flag if set
			flags &= ~FLAG_USE_CFG_FILE;
				std::cout << "'-u' is being overridden by '-i'. Ignoring " << DEFAULT_OUT_NAME << "." << std::endl;
		}
	} // if (flags & FLAG_USE_CFG_FILE)
}

void processFiles(std::string ifileName, int& flags)
{
	if (ifileName.empty() == true)
	{
		std::string err = "Empty parameter passed.";
		reportErr("processFiles", err, -1, flags);
	}
	
	std::cout << "processing " << ifileName << std::endl;
	
	if (flags & FLAG_VERBOSE)
	{
		std::cout << "processFiles: just processing " << ifileName << std::endl;
	}
	
	std::string::size_type pos;

	// search for * wild card
	pos = ifileName.find("*");
	if (pos != std::string::npos)
	{
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "processFiles: wild card found!" << std::endl;
		}

		std::string ext = getFileExt(ifileName, flags);

		if (ext.empty())
		{
			std::string err = "The filename - '"; err += ifileName; err += "' has wildcard but no extension.";
			reportErr("processFiles", err, -1, flags);
		}
		
		// setup cfg
		pro_ext.push_back(ext);
		
		if (flags & FLAG_VERBOSE)
		{
			std::cout << "processFiles: updating flags" << std::endl;
		}

		// as we have setup cfg, use it
		flags |= FLAG_USE_CFG_FILE;
		flags |= FLAG_PROCESS_EXT;
		// we will now on behave as if ext was defined in cfg
		flags &= ~FLAG_USE_ARGS;
		
		validateFlags(flags);
		
		processDirs(flags);
	}
	else // if (pos != std::string::npos)
	{
		klocfile(ifileName, flags);
	} // if (pos != std::string::npos)
}

void reportErr(std::string fnName, std::string msg, int exitStat, int flags)
{
	if (flags & FLAG_VERBOSE)
	{
		std::cerr << fnName << ": ";
	}

	std::cerr << msg << std::endl;
	
	if (exitStat < 0)
	{
		if (flags & FLAG_VERBOSE)
		{
			std::cerr << fnName << ": errno reported was - " << errno << std::endl;
			std::cerr << fnName << ": ";
		}
		
		std::cerr << "Confused by earlier error, bailing out." << std::endl;
		
		exit (exitStat);
	}
	else
	{
		if (flags & FLAG_VERBOSE)
		{
			std::cerr << fnName << ": errno reported was - " << errno << std::endl;
			std::cerr << fnName << ": ";
		}
		
		std::cerr << "Recovered from earlier error, continuing processing." << std::endl;
	}
	
}
	
void usage(int flags)
{
	std::cout << "Usage: kloc [OPTION]" << std::endl;
	std::cout << "This is kloc " << VERSION << ", a source code statics program." << std::endl;
	std::cout << "" << std::endl;
	std::cout << "  -d      process the current working dir" << std::endl;
	std::cout << "  -g      generate only grouped data" << std::endl;
	std::cout << "  -f      do not recurse" << std::endl;
	std::cout << "  -s      generate more stats" << std::endl;
	std::cout << "  -c      generate configuration file skeleton and exit" << std::endl;
	std::cout << "  -u      use the configuration file" << std::endl;
	std::cout << "  -i file process only specified file" << std::endl;
	std::cout << "  -a      ignore hidden files (starting with a period)" << std::endl;
	std::cout << "  -v      be verbose" << std::endl;
	std::cout << "  -h      display this brief help and exit" << std::endl;
	std::cout << "  -H      display very detailed help and exit" << std::endl;
	std::cout << std::endl;
	std::cout << "Report bugs to ddiinnxx at users.sarovar.org" << std::endl;

	return;
}

void more_usage(int flags)
{
	usage(flags);
	
	std::cout << std::endl;
	std::cout << "Configuring kloc" << std::endl;
	std::cout << "================" << std::endl;
	std::cout << std::endl;
	std::cout << "The " << DEFAULT_OUT_NAME << " can be used to process or ignore specific set of file(s)." << std::endl;
	std::cout << std::endl;
	std::cout << "kloc.cfg had four fields. Process File, Process Extension, Ignore File" << std::endl;
	std::cout << "and Ignore Extension." << std::endl;
	std::cout << std::endl;
	std::cout << "kloc will process files with extensions specified in [Process Extension]. An " << std::endl;
	std::cout << "empty [Process Extension] will cause all files to be processed. " << std::endl;
	std::cout << std::endl;
	std::cout << "kloc will also process files specified in [Process File]. It is particularly " << std::endl;
	std::cout << "useful for files with no extension. " << std::endl;
	std::cout << std::endl;
	std::cout << "If both [Process Extension] and [Process File] have data, kloc will process " << std::endl;
	std::cout << "the union of the two sets." << std::endl;
	std::cout << std::endl;
	std::cout << "kloc will ignore files with extensions specified in [Ignore Extension]. An " << std::endl;
	std::cout << "empty [Ignore Extension] will cause no files to be ignored. Entries in " << std::endl;
	std::cout << "[Process Extension] will over-ride." << std::endl;
	std::cout << std::endl;
	std::cout << "kloc will also ignore files specified in [Ignore File]. It is particularly " << std::endl;
	std::cout << "useful for files with no extension. Entries in [Process File] will over-ride." << std::endl;
	std::cout << std::endl;
	std::cout << "If both [Ignore Extension] and [Ignore File] have data, kloc will ignore the " << std::endl;
	std::cout << "union of the two sets." << std::endl;
	std::cout << std::endl;
	std::cout << "The results of the various combinations is given below." << std::endl;
	std::cout << std::endl;
	std::cout << "	if    |    then" << std::endl;
	std::cout << "------------------------" << std::endl;
	std::cout << "     PF   PE  |  IF   IE" << std::endl;
	std::cout << "i    1    1   |  x    0" << std::endl;
	std::cout << "ii   1    0   |  0    0" << std::endl;
	std::cout << "iii  0    1   |  x    0" << std::endl;
	std::cout << "iv   0    0   |  x    x" << std::endl;
	std::cout << "" << std::endl;
	std::cout << "PF - Process File" << std::endl;
	std::cout << "PE - Process Extension" << std::endl;
	std::cout << "IF - Ignore File" << std::endl;
	std::cout << "IE - Ignore Extension" << std::endl;
	std::cout << "1  - Will be considered" << std::endl;
	std::cout << "0  - Will not be considered even if entry present" << std::endl;
	std::cout << "x  - Will depend if entry is present" << std::endl;
	std::cout << "" << std::endl;
	std::cout << "How it works" << std::endl;
	std::cout << "================" << std::endl;
	std::cout << "kloc, as of now, recognises //, # and /* */ as comment markers. If //, # or" << std::endl;
	std::cout << "/* found at begining of a line, the comment counter is incremented and if" << std::endl;
	std::cout << "found in the middle of a line, the embedded counter is incremented." << std::endl;
	std::cout << "" << std::endl;
	std::cout << "By dividing Total Empty Lines by Total Lines, we get how much % of written lines" << std::endl;
	std::cout << "were code." << std::endl;
	std::cout << "" << std::endl;
	std::cout << "By dividing sum of Total Comments and Total Embedded Comments by Total Code, we" << std::endl;
	std::cout << "get how much % of code was commented." << std::endl;
	std::cout << "" << std::endl;
	std::cout << "By dividing of sum of Total Comments, Total Embedded Comments and Total Code by" << std::endl;
	std::cout << "Total Lines we get how much % of written lines were code or comments." << std::endl;
	std::cout << std::endl;
	std::cout << "Typical Usage" << std::endl;
	std::cout << "================" << std::endl;
	std::cout << std::endl;
	std::cout << "$ kloc -gs -i ./kloc.cpp" << std::endl;
	std::cout << "$ kloc -gs -i \"*.h\"" << std::endl;
	std::cout << "$ kloc -c" << std::endl;
	std::cout << "$ kloc -uv" << std::endl;
	std::cout << "$ kloc -g" << std::endl;
	std::cout << "" << std::endl;

	return;
}

/*
$Date: 2006/09/22 06:37:32 $
$Revision: 1.8 $

history
---------
3i,21sep06,dc,1.8  fixed bug (reported by chris deMots) where closedir(dirn) 
                   called even when dirn was NULL causing seg fault in solaris.
                   added check for return value of closedir(), used tabs
3h,07sep06,dc,1.7  fixed bug (reported by nick) where DIR handle was not closed
                   leading to too many descriptor in large dir heirarchy, fixed
                   bug in VERSION usage in cfg file, used tabs
3g,21feb06,dc,1.6  corrected bug where code line count was not incremented when
                   there was an embedded comment
3f,15oct05,dc,1.5  used version from VERSION
3e,09oct05,dc,1.4  used tabs, updated help, added verbose comments
3d,09oct05,dc      fixed bug in ignore file in cfg file
3c,09oct05,dc      added code (sent by martin sigrand) for supporting c-style 
                   comments / * * /
3b,18jun04,dc      added code (sent by larry sinn) for ignoring hidden files
3a,19sep03,dc      first stable version
2b,08jan03,dc      first public release via dinkercharak.com
1a,17dec02,dc      written
*/
