//****************************************************************************
// Copyright (C) 2005,  Lucian Constantin <struct_bylighting@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//****************************************************************************

#include "qwerty.h"
#include "briefedit.h"
#include "dialogs.h"
#include "lineandcolumn.h"
#include "bufmgr.h"
#include <qapplication.h>

void Editor::undo() {
	cleanMWords();
	e_->undo();
}

void Editor::cut() {
	LOG("Editor cut");
	cleanMWords();
	e_->cut();
}
void Editor::copy() {
	cleanMWords();

	// try first to copy from the current tab window
	ASSERT(tw_current_ != 0);
	QString text = tw_current_->markedText();
	if (!(text.isNull() || text.isEmpty()))
		tw_current_->copy();
	
	// otherwise copy from edit window
	else
		e_->copy();
}

void Editor::paste() {
	cleanMWords();
	e_->paste();
}
void Editor::e_delete() {
	cleanMWords();
	e_->del();
}

void Editor::selectAll() {
	cleanMWords();
	e_->selectAll(true);
}

void Editor::toggleOverwrite() {
	bool status = e_->isOverwriteMode();
LOG("%d overwrite");
	e_->setOverwriteMode(!status);
	
	if (status)
		iow_->setText("Insert");
	else
		iow_->setText("Overwrite");
}

void Editor::startOfBuffer() {
	cleanMWords();
	e_->setCursorPosition(0, 0);
}

void Editor::endOfBuffer() {
	cleanMWords();
	int lines = e_->numLines();
	e_->setCursorPosition(lines - 1, 0);
}

void Editor::gotoLine() {
	cleanMWords();
	LineDialog *ld = new LineDialog();
	ld->setCaption("Enter line number");
	if (!ld->exec())
		return;

	int line = ld->getLine();
	LOG("Editor::gotoline %d", line);
	delete ld;

	if (line > 0) {
		int lines = e_->numLines();
		if (line > lines)
			line = lines;
		e_->setCursorPosition(line - 1, 0);
	}
}

void Editor::gotoBookmark() {
	cleanMWords();
	BookmarkDialog *bd = new BookmarkDialog();
	bd->setCaption("Enter bookmark number");
	if (!bd->exec())
		return;
	int bookmark = bd->getBookmark();
	LOG("Editor::gotoBookmark %d", bookmark);
	delete bd;

	gotoBookmarkCommon(bookmark);
}

void Editor::gotoBookmarkCommon(int bookmark) {
	if (bookmark > MAXBOOKMARKS || bookmark < 0)
		return;

	// change the buffer if necessary
	QString current_fname = getFname();
	if (current_fname != book_[bookmark].fname_) {
		bool found = false;
		int i;
		for (i = 0; i < MAXBUFMGR; i++) {
			QString bfname = b_->getFName(i);
			if (bfname.isNull())
				break;
			if (bfname.startsWith("*"))
				bfname = bfname.mid(1);
			if (book_[bookmark].fname_ == bfname) {
				found = true;
				break;
			}
		}
		if (!found)
			return;
		// save current buffer data
		saveCurentBuffer();
	
		// move to the buffer specified by the bookmark
		b_->setActiveIndex(i);
	
		// restore data
		if (using_aux_)
			changed_aux_ = true;
		else
			changed_aux_ = true;
		QString fname = b_->getActiveFName();
		setFname(fname, b_->getActiveFormat(), false);
		e_->setText(b_->getActiveText());	
		if (using_aux_)
			changed_aux_ = fname.startsWith("*");
		else
			changed_main_ = fname.startsWith("*");
		int line = b_->getActiveLine();
		int col = b_->getActiveCol();
		e_->setCursorPosition(line, col, false);

		redoBufMgrList();
	}


	int line = book_[bookmark].y_;
	int lines = e_->numLines();
	if (line > lines)
		line = lines - 1;
	e_->setCursorPosition(line, book_[bookmark].x_);
	e_->setFocus();
}

void Editor::setBookmark() {
	cleanMWords();
	BookmarkDialog *bd = new BookmarkDialog();
	bd->setCaption("Enter bookmark number");
	if (!bd->exec())
		return;
	int bookmark = bd->getBookmark();
	LOG("Editor::setBookmark %d", bookmark);
	delete bd;

	setBookmarkCommon(bookmark);
}

void Editor::regenBookmarks() {
	// regenerate bookmark window
	QString book = "";
	for (int i = 0; i < MAXBOOKMARKS; i++) {
		if (!book_[i].fname_.isEmpty()) {
			char str[20];
			sprintf(str, "%d: ", i);
			book += str + book_[i].fname_ + ":\n\t" + book_[i].line_ + "\n";
		}
	}

	tw_bookmarks_->setText(book);
	tw_->setCurrentPage(3); //lucian: hardcoded
}

void Editor::setBookmarkCommon(int bookmark) {
	cleanMWords();
	if (bookmark > MAXBOOKMARKS)
		return;

	int para;
	int index;
	e_->getCursorPosition(&para, &index);
	book_[bookmark].x_ = index;
	book_[bookmark].y_ = para;
	book_[bookmark].fname_ = getFname();
	book_[bookmark].line_ = e_->textLine(para);
	regenBookmarks();
	
}

void Editor::setBookmark1() {
	setBookmarkCommon(1);
}

void Editor::setBookmark2() {
	setBookmarkCommon(2);
}

void Editor::setBookmark3() {
	setBookmarkCommon(3);
}

void Editor::setBookmark4() {
	setBookmarkCommon(4);
}

void Editor::setBookmark5() {
	setBookmarkCommon(5);
}

void Editor::setBookmark6() {
	setBookmarkCommon(6);
}

void Editor::setBookmark7() {
	setBookmarkCommon(7);
}

void Editor::setBookmark8() {
	setBookmarkCommon(8);
}

void Editor::setBookmark9() {
	setBookmarkCommon(9);
}

void Editor::setBookmark0() {
	setBookmarkCommon(0);
}

void Editor::insertHex() {
	cleanMWords();
	HexDialog *hd = new HexDialog();
	hd->setCaption("Enter unicode code");
	if (!hd->exec())
		return;
	int val = hd->getHex();
	if (val == -1)
		return;

	int para;
	int pos;
	e_->getCursorPosition(&para, &pos);
	QString newstr = QChar((unsigned short) val);
	e_->insertAt(newstr, para, pos);
	LOG("Editor::insertHex %x", val);
	delete hd;
}
