//****************************************************************************
// Copyright (C) 2005,  Lucian Constantin <struct_bylighting@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//****************************************************************************

#include "toc.h"

// simple C/C++ parser
//
// state0 - start new text
//
// state0 ---;---> state1 - we have a full expression "int a;" or "int a = 7;"; also working for "typedef unsigned UINT32;"
//
// state0 ---{----> state2 - starting expression "int a = {"
// state2 ---{----> state2 - do {} matching
// state2 ---}----> state3 
// state3 ---;----> state4 - we have a full expression "int a = {};"; also working for "typedef struct list {struct list *next; int val;} LIST, *pLIST;"
//
// state0 ---(----> state5 - starting to match a function
// state5 ---)----> state6
// state6 ---;----> state0 - this was just a function declaration "int f();"; do not add it to results
// state6 ---{----> state7
// state7 ---{----> state7 - do {} matching
// state7 ---}----> state8 - we have a full ANSI function definition
// state6 ---any char but not ;---->state10 - we are dealing here with a non-ANSI function definition
// state10 ---{---> state11
// state11 ---{---> state11 - do {} matching
// state11 ---}---> state12 - we have a non-ANSI function definition
//
// state0 ---#---> state9 - we do not care for macros
// state9 ---\---> state9 - advance one line
// state9 ---end of string---->state 0
QString TocC::parse(Q3MultiLineEdit *e) {
	int state = 0; // state0
	int matching = 0;
	bool advance_line = false;
	bool in_comment = false;
	int para_statement = 0;
	QString line_statement = "";
	QString retval = "";
	int para = 0;
	int lines = e->numLines();

	if (lines == 0)
		return (QString) "";

	while (para < lines) {
		QString line = e->textLine(para);
		if (state == 0) {
			para_statement = para;
			line_statement = line;
		}
		const char *ptr = (const char *) line;
		while (*ptr != '\0') {
			// first eliminate comments
			if (!in_comment && *ptr == '/' && *(ptr +1) == '/') { // C++ comment
				break; // exit from while (*ptr != '\0')
			}
			else if (!in_comment && *ptr == '/' && *(ptr +1) == '*') { // C comment
//printf("start comment on line %d\n", para + 1);
				in_comment = true;
				ptr += 2;
			}
			
			if (in_comment) {
//printf("in comment: %d:%s\n", para + 1, ptr); 
				bool found = false;
				while (*ptr != '\0') {
					if (*ptr == '*' && *(ptr + 1) == '/') {
						found = true;
						ptr += 2;
						in_comment = false;
//printf("end comment on line %d\n", para + 1);
						break;
					}
					ptr++;
				}
				if (!found) {
					break; // exit from while (*ptr != '\0')
				}
			}
			
			// run the state machine
			switch (state) {
				case 0:
					if (*ptr == ';')
						state = 1;
					else if (*ptr == '{')
						state = 2;
					else if (*ptr == '(')
						state = 5;
					else if (*ptr == '#')
						state = 9;
					break;
				case 2:
					if (*ptr == '{')
						matching++;
					else if (*ptr == '}' && matching != 0)
						matching--;
					else if (*ptr == '}')
						state = 3;
					break;
				case 3:
					if (*ptr == ';')
						state = 4;
					break;
				case 5:
					if (*ptr == ')')
						state = 6;
					break;
				case 6:
					if (*ptr == ';')
						state = 0;
					else if (*ptr == '{')
						state = 7;
					else if (*ptr != ' ' && *ptr != '\t')
						state = 10;
					break;
				case 7:
					if (*ptr == '{')
						matching++;
					else if (*ptr == '}' && matching != 0)
						matching--;
					else if (*ptr == '}')
						state = 8;
					break;
				case 9:
					if (*ptr == '\\')
						advance_line = true;
					break;
				case 10:
					if (*ptr == '{')
						state = 11;
					break;
				case 11:
					if (*ptr == '{')
						matching++;
					else if (*ptr == '}' && matching != 0)
						matching--;
					else if (*ptr == '}')
						state = 12;
					break;
				default:
					break;
			};
			
			// advance to a new line
			if (advance_line) {
				break; // exit from while (*ptr != '\0')
			}
			
			// look for termination states
			if (state == 1 || state == 4 || state == 8 || state == 12) {
				char str[50];
				sprintf(str, "%d: ", para_statement + 1);
				retval += str + line_statement + "\n";
				state = 0;
				matching = 0;
			}
		
			ptr++;
		}

		// skeep macros
		if (state == 9 && !advance_line)
			state = 0;
		
		advance_line = false;
		para++;
	}
	
	return retval;
}

// simple LaTeX parser: just print the toc based on
// \chapter{}, \section{}, and \subsection{}
QString TocLatex::parse(Q3MultiLineEdit *e) {
	QString retval = "";
	int para = 0;
	int lines = e->numLines();

	if (lines == 0)
		return (QString) "";

	while (para < lines) {
		QString line = e->textLine(para);
		const char *str = (const char *) line;
		if (strstr(str, "\\chapter") != 0 ||
		    strstr(str, "\\section") != 0 ||
		    strstr(str, "\\subsection") != 0) {
			char str[50];
			sprintf(str, "%d: ", para + 1);
			retval += str + line + "\n";
		}
		para++;
	}
	
	return retval;
}
